#include <stdlib.h>
#include <stdio.h>

#include <sys/types.h>	// For _ftime() and struct _timeb.
#include <sys/timeb.h>	// For ftime(), _ftime(), struct timeb, struct _timeb.
#include <time.h>		// For clock() and CLOCKS_PER_SEC.

#include "timing.h"


/////////////////////////////////////////////////////////////////////////////
// Define the function gettimeofday() if system is WIN32.
/////////////////////////////////////////////////////////////////////////////
#ifdef _WIN32

#include <time.h>
#include <windows.h>

#if defined(_MSC_VER) || defined(_MSC_EXTENSIONS)
  #define DELTA_EPOCH_IN_MICROSECS  11644473600000000Ui64
#else
  #define DELTA_EPOCH_IN_MICROSECS  11644473600000000ULL
#endif


struct timezone 
{
  int  tz_minuteswest; /* minutes W of Greenwich */
  int  tz_dsttime;     /* type of dst correction */
};


static int gettimeofday(struct timeval *tv, struct timezone *tz)
{
	FILETIME ft;
	unsigned __int64 tmpres = 0;
	static int tzflag = 0;

	if (NULL != tv)
	{
		GetSystemTimeAsFileTime(&ft);

		tmpres |= ft.dwHighDateTime;
		tmpres <<= 32;
		tmpres |= ft.dwLowDateTime;

		tmpres /= 10;  /*convert into microseconds*/
		/*converting file time to unix epoch*/
		tmpres -= DELTA_EPOCH_IN_MICROSECS; 
		tv->tv_sec = (long)(tmpres / 1000000UL);
		tv->tv_usec = (long)(tmpres % 1000000UL);
	}

	if (NULL != tz)
	{
		if (!tzflag)
		{
			_tzset();
			tzflag++;
		}
		tz->tz_minuteswest = _timezone / 60;
		tz->tz_dsttime = _daylight;
	}

	return 0;
}
#else

#include <sys/time.h>	// For gettimeofday() and struct timeval.

#endif



/////////////////////////////////////////////////////////////////////////////
// Returns time in seconds (plus fraction of a second) since 
// midnight (00:00:00), January 1, 1970, coordinated universal time (UTC).
// Up to microsecond precision.
/////////////////////////////////////////////////////////////////////////////
double GetCurrRealTime_A( void )
{
	struct timeval tm;
    gettimeofday( &tm, NULL );
	return ((double)tm.tv_sec + ((double)tm.tv_usec / 1000000.0));
}



/////////////////////////////////////////////////////////////////////////////
// Returns time in seconds (plus fraction of a second) since 
// midnight (00:00:00), January 1, 1970, coordinated universal time (UTC).
// Up to millisecond precision.
/////////////////////////////////////////////////////////////////////////////
double GetCurrRealTime_B( void )
{
#ifdef _WIN32
	struct _timeb timebuffer;
	_ftime( &timebuffer );
#else
	struct timeb timebuffer;
	ftime( &timebuffer );
#endif
	return ((double)timebuffer.time + ((double)timebuffer.millitm / 1000.0));
}



/////////////////////////////////////////////////////////////////////////////
// Returns CPU time in seconds (plus fraction of a second) since the 
// start of the current process.
/////////////////////////////////////////////////////////////////////////////
double GetCurrCPUTime( void )
{
	return ((double) clock() ) / CLOCKS_PER_SEC;
}

