//============================================================
// STUDENT NAME: <your name>
// MATRIC NO.  : <matric no.>
// NUS EMAIL   : <your NUS email address>
// COMMENTS TO GRADER:
// <comments to grader, if any>
//
// ============================================================
//
// FILE: ripplemap.frag


const float pi = 3.14159265;


// Eye-space position and vectors for setting up a tangent space at the fragment.

varying vec3 ecPosition; // Fragment's 3D position in eye space.
varying vec3 ecNormal;	 // Fragment's normal vector in eye space.
varying vec3 ecTangent;  // Frgament's tangent vector in eye space.


// Sine wave has equation y = A * sin( 2*pi*f*x - p ).

uniform float Amplitude; // A
uniform float Frequency; // f
uniform float Phase;     // p


// The environment cubemap for reflection mapping.

uniform samplerCube EnvMap; 


// The final fragment color is obtained by linearly interpolating between the 
// color computed using Phong lighting model and the color from the environment map.
// When EnvBlend is 0.0, the result is the Phong lighting color, and
// when EnvBlend is 1.0, the result is the environment map color.

uniform float EnvBlend; 



void main()
{
    vec3 N = normalize( ecNormal );
    vec3 B = normalize( cross( N, ecTangent ) );
    vec3 T = cross( B, N );

    vec3 ecViewVec = -normalize( ecPosition );
    vec3 ecLightPos = vec3( gl_LightSource[0].position ) / gl_LightSource[0].position.w;
    vec3 ecLightVec = normalize( ecLightPos - ecPosition );

    vec3 tanLightVec;
    tanLightVec.x = dot( ecLightVec, T );
    tanLightVec.y = dot( ecLightVec, B );
    tanLightVec.z = dot( ecLightVec, N );
    tanLightVec = normalize( tanLightVec );

    vec3 tanViewVec;
    tanViewVec.x = dot( ecViewVec, T );
    tanViewVec.y = dot( ecViewVec, B );
    tanViewVec.z = dot( ecViewVec, N );
    tanViewVec = normalize( tanViewVec );
    
    vec2 dir = vec2( 0.5 ) - fract( gl_TexCoord[0].st );
    float x = length( dir );
    float omega = 2.0 * pi * Frequency;
    float dydx = omega * Amplitude * cos( omega * x - Phase );
    
    vec3 tanNormDelta = normalize( vec3( dydx * dir.x, dydx * dir.y, x ) );

    vec3 tanHalfVector = normalize( tanLightVec + tanViewVec );

    float N_dot_L = max( 0.0, dot( tanNormDelta, tanLightVec ) );
    float N_dot_H = max( 0.0, dot( tanNormDelta, tanHalfVector ) );

    float pf = ( N_dot_H == 0.0 )? 0.0 : pow( N_dot_H, gl_FrontMaterial.shininess );
    
    vec4 phongColor = gl_FrontLightModelProduct.sceneColor +
                      gl_LightSource[0].ambient * gl_FrontMaterial.ambient +
                      gl_LightSource[0].diffuse * gl_FrontMaterial.diffuse * N_dot_L +
                      gl_LightSource[0].specular * gl_FrontMaterial.specular * pf;

    vec3 tanReflectVec = reflect( -tanViewVec, tanNormDelta );
    vec3 ecReflectVec = tanReflectVec.x * T + tanReflectVec.y * B + tanReflectVec.z * N;
    vec4 reflectionColor = textureCube( EnvMap, ecReflectVec );

    gl_FragColor = mix( phongColor, reflectionColor, EnvBlend );
}
